from typing import Hashable, List, Sequence, Tuple, Type, TypeVar

import networkx as nx
from jax.tree_util import register_pytree_node

from .annotations import PyTree

__all__: List[str] = []

T = TypeVar('T', bound=nx.Graph)


def register_graph_as_jax_pytree(cls: Type[T]) -> None:
    def tree_unflatten(hashed: Hashable, trees: Sequence[PyTree]) -> T:
        node_dicts, edge_dicts = trees

        if not isinstance(node_dicts, dict):
            raise TypeError
        if not isinstance(edge_dicts, dict):
            raise TypeError

        graph = cls()
        graph.add_nodes_from(node_dicts.items())
        graph.add_edges_from([(source, target, data)
                              for (source, target), data in edge_dicts.items()])

        return graph

    def tree_flatten(self: T) -> Tuple[Sequence[PyTree], Hashable]:
        return ((dict(self.nodes), dict(self.edges)), None)

    register_pytree_node(cls, tree_flatten, tree_unflatten)


register_graph_as_jax_pytree(nx.Graph)
register_graph_as_jax_pytree(nx.DiGraph)



from collections.abc import Mapping
from functools import singledispatch
from numbers import Number
from typing import Any, List, Tuple, Type, Union

import colorful as cf
import networkx as nx
import numpy as np
from jax.errors import TracerArrayConversionError
from jax.interpreters.ad import JVPTracer
from jax.interpreters.batching import BatchTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer, JaxprTracer
from jax.interpreters.xla import DeviceArray

from .annotations import Array

__all__ = ['print_generic', 'display_generic']


# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/2904.


def print_generic(*args: Any, **kwargs: Any) -> None:
    for value in args:
        print(display_generic(value))
    for key, value in kwargs.items():
        print(display_key_and_value(key, value, "=", True, 0))


@singledispatch
def display_generic(value: Any, show_values: bool = True, indent: int = 0) -> str:
    return cf.red(str(value)) + "\n"


@display_generic.register
def _(value: JVPTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"JVPTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register  # type: ignore
def _(value: JaxprTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"JaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register  # type: ignore
def _(value: DynamicJaxprTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"DynamicJaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register  # type: ignore
def _(value: BatchTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"BatchTracer {value.shape} {value.dtype} "
                      f"batched over {value.val.shape[value.batch_dim]}") + "\n"


@display_generic.register(np.ndarray)  # type: ignore
def _(value: Array, show_values: bool = True, indent: int = 0) -> str:
    retval = cf.yellow(f"NumPy Array {value.shape} {value.dtype}") + "\n"
    return retval + _show_array(indent + 1, value)


@display_generic.register  # type: ignore
def _(value: DeviceArray, show_values: bool = True, indent: int = 0) -> str:
    base_string = cf.violet(f"Jax Array {value.shape} {value.dtype}") + "\n"
    try:
        np_value = np.asarray(value)
    except TracerArrayConversionError:
        array_string = ""
    else:
        array_string = _show_array(indent + 1, np_value)
    return base_string + array_string


@display_generic.register(type(None))  # type: ignore
@display_generic.register(Number)
def _(value: Union[None, Number], show_values: bool = True, indent: int = 0) -> str:
    return cf.cyan(str(value)) + "\n"


@display_generic.register(Mapping)  # type: ignore
def _(value: Mapping[Any, Any], show_values: bool = True, indent: int = 0) -> str:
    return (display_class(type(value))
            + "".join(display_key_and_value(key, sub_value, "=", show_values, indent)
                      for key, sub_value in sorted(value.items())))


@display_generic.register(tuple)  # type: ignore
@display_generic.register(list)
def _(value: Union[Tuple[Any, ...], List[Any]], show_values: bool = True, indent: int = 0) -> str:
    return (display_class(type(value))
            + "".join(display_key_and_value("", sub_value, "", show_values, indent)
                      for sub_value in value))


@display_generic.register  # type: ignore
def _(value: nx.Graph, show_values: bool, indent: int = 0) -> str:
    directed = isinstance(value, nx.DiGraph)
    arrow = cf.base00('⟶  ' if directed else '🡘 ')
    retval = display_class(type(value))
    for name, node in value.nodes.items():
        retval += display_key_and_value(name, node, ": ", show_values, indent)
    for (source, target), edge in value.edges.items():
        key = f"{source}{arrow}{target}"
        retval += display_key_and_value(key, edge, ": ", show_values, indent)
    return retval


def display_class(cls: Type[Any]) -> str:
    return cf.orange(cls.__name__) + "\n"


def display_key_and_value(
        key: str, value: Any, separator: str, show_values: bool = True, indent: int = 0) -> str:
    return (_indent_space(indent + 1) + cf.blue(key) + cf.base00(separator)
            + display_generic(value, show_values, indent + 1))


# Private functions --------------------------------------------------------------------------------
_indentation = 4


def _indent_space(indent: int) -> str:
    return (indent * _indentation) * " "


def _format_number(x: float) -> str:
    return f"{x:10.4f}"


def _show_array(indent: int, array: Array) -> str:
    if len(array.shape) == 0:
        return _indent_space(indent) + _format_number(array[()]) + "\n"
    if np.prod(array.shape) == 0:
        return ""
    if 1 in array.shape:
        return _show_array(indent, array[tuple(0 if s == 1 else slice(None)
                                               for s in array.shape)])
    if any(x > 10 for x in array.shape):
        return ""
    if len(array.shape) == 1:
        return _indent_space(indent) + "  ".join(_format_number(array[i])
                                                 for i in range(array.shape[0])) + "\n"
    if len(array.shape) == 2:
        return "".join(_show_array(indent, array_slice)
                       for array_slice in array)
    return ""



from .annotations import *
from .color_stub import *
from .cotangent_tools import *
from .dataclasses import *
from .display import *
from .dtypes import *
from .generator import *
from .graph import *
from .leaky_integral import *
from .partial import *
from .plottable_trajectory import *
from .shims import *
from .testing import *
from .tools import *



from __future__ import annotations

from numbers import Integral
from typing import Optional, Tuple, overload

import jax.numpy as jnp
import numpy as np
from jax.lax import scan

from .annotations import ComplexArray, ComplexNumeric, RealArray, RealNumeric
from .dataclasses import dataclass
from .dtypes import real_dtype
from .generator import Generator

__all__ = ['leaky_integrate', 'diffused_leaky_integrate', 'leaky_data_weight',
           'leaky_integrate_time_series', 'leaky_covariance']


@overload
def leaky_integrate(value: RealNumeric,
                    time_step: RealNumeric,
                    drift: Optional[RealNumeric] = None,
                    decay: Optional[RealNumeric] = None,
                    *,
                    leaky_average: bool = False) -> RealNumeric:
    ...


@overload
def leaky_integrate(value: ComplexNumeric,
                    time_step: RealNumeric,
                    drift: Optional[ComplexNumeric] = None,
                    decay: Optional[ComplexNumeric] = None,
                    *,
                    leaky_average: bool = False) -> ComplexNumeric:
    ...


def leaky_integrate(value: ComplexNumeric,
                    time_step: RealNumeric,
                    drift: Optional[ComplexNumeric] = None,
                    decay: Optional[ComplexNumeric] = None,
                    *,
                    leaky_average: bool = False) -> ComplexNumeric:
    """
    Update the value so that it is the leaky integral (or leaky average).
    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        decay: If provided, must have positive real component, and the value decays by exp(-decay)
            every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    if drift is None:
        if decay is None:
            return value
        return value * jnp.exp(-decay * time_step)

    if decay is None:
        if leaky_average:
            raise ValueError
        return value + drift * time_step

    scaled_integrand = (drift / decay) * -jnp.expm1(-decay * time_step)

    if leaky_average:
        scaled_integrand *= decay.real

    return value * jnp.exp(-decay * time_step) + scaled_integrand


@overload
def diffused_leaky_integrate(value: RealArray,
                             time_step: RealNumeric,
                             rng: Generator,
                             diffusion: RealNumeric,
                             drift: Optional[RealNumeric] = None,
                             decay: Optional[RealNumeric] = None,
                             *,
                             leaky_average: bool = False) -> RealNumeric:
    ...


@overload
def diffused_leaky_integrate(value: ComplexArray,
                             time_step: RealNumeric,
                             rng: Generator,
                             diffusion: RealNumeric,
                             drift: Optional[ComplexNumeric] = None,
                             decay: Optional[ComplexNumeric] = None,
                             *,
                             leaky_average: bool = False) -> ComplexNumeric:
    ...


def diffused_leaky_integrate(value: ComplexArray,
                             time_step: RealNumeric,
                             rng: Generator,
                             diffusion: RealNumeric,
                             drift: Optional[ComplexNumeric] = None,
                             decay: Optional[ComplexNumeric] = None,
                             *,
                             leaky_average: bool = False) -> ComplexNumeric:
    """
    Update an Ornstein-Uhlenbeck process.

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        rng: The random number generator for the stochastic process.
        diffusion: The diffusion for the stochastic process.
        decay: If provided, the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    variance = (diffusion * time_step
                if decay is None
                else diffusion / decay * -jnp.expm1(-decay * time_step))
    jump = jnp.sqrt(variance) * rng.normal(value.shape)
    return leaky_integrate(value, time_step, drift, decay, leaky_average=leaky_average) + jump


def leaky_data_weight(iterations_times_time_step: RealNumeric,
                      decay: RealNumeric) -> RealNumeric:
    """
    Returns: The amount of data that has been incorporated and has not been decayed.  That is,
        leaky_integrate(0.0, iterations_times_time_step, 1.0, decay, leaky_average=True)
    """
    return -jnp.expm1(-iterations_times_time_step * decay)


@dataclass
class _FilterCarry:
    iterations: RealNumeric
    value: ComplexArray


def leaky_integrate_time_series(time_series: ComplexArray, decay: ComplexNumeric) -> ComplexArray:

    def g(carry: _FilterCarry, drift: ComplexNumeric) -> Tuple[_FilterCarry, ComplexArray]:
        new_iterations = carry.iterations + 1.0
        data_weight = leaky_data_weight(new_iterations, decay.real)  # type: ignore

        new_value = leaky_integrate(carry.value, 1.0, drift, decay, leaky_average=True)
        assert isinstance(new_value, jnp.ndarray)
        new_carry = _FilterCarry(new_iterations, new_value)
        outputted_value = new_value / data_weight
        return new_carry, outputted_value

    # Cast the dtype from integer to floating point to prevent integer rounding.
    initial_value = jnp.zeros(time_series[0].shape,
                              dtype=(real_dtype()
                                     if issubclass(time_series.dtype.type, Integral)
                                     else time_series.dtype))
    initial_carry = _FilterCarry(0.0, initial_value)

    _, filtered_time_series = scan(g, initial_carry, time_series)
    return filtered_time_series


def leaky_covariance(x_time_series: ComplexArray,
                     y_time_series: ComplexArray,
                     decay: ComplexNumeric,
                     covariance_matrix: bool = False) -> ComplexArray:
    if covariance_matrix:
        if x_time_series.shape[0] != y_time_series.shape[0]:
            raise ValueError
        s = (np.newaxis,)

        def times(a: ComplexArray, b: ComplexArray) -> ComplexArray:
            return a[(..., *(s * (b.ndim - 1)))] * b[(slice(None), *(s * (a.ndim - 1)))]
    else:
        if x_time_series.shape != y_time_series.shape:
            raise ValueError
        times = jnp.multiply
    x = leaky_integrate_time_series(x_time_series, decay)
    y = leaky_integrate_time_series(y_time_series, decay)
    xy = leaky_integrate_time_series(times(x_time_series, y_time_series), decay)
    return xy - times(x, y)



from __future__ import annotations

from functools import partial
from typing import Generic, Tuple

import jax.numpy as jnp
from jax.tree_util import tree_map, tree_multimap, tree_reduce

from ..annotations import BooleanNumeric, ComplexNumeric, RealNumeric
from ..dataclasses import dataclass
from ..leaky_integral import leaky_data_weight, leaky_integrate
from ..tools import abs_square, divide_nonnegative
from .augmented import AugmentedState, State
from .combinator import Differentiand, IteratedFunctionWithCombinator
from .iterated_function import Comparand, IteratedFunction, Parameters, Trajectory

__all__ = ['StochasticState', 'StochasticIteratedFunction',
           'StochasticIteratedFunctionWithCombinator']


@dataclass
class StochasticState(AugmentedState[State], Generic[State, Comparand]):
    mean_state: Comparand
    second_moment_state: Comparand


@dataclass
class StochasticIteratedFunction(
        IteratedFunction[Parameters, State, Comparand, Trajectory,
                         StochasticState[State, Comparand]],
        Generic[Parameters, State, Comparand, Trajectory]):
    convergence_detection_decay: RealNumeric

    # Implemented methods --------------------------------------------------------------------------
    def initial_augmented(self, initial_state: State) -> StochasticState[State, Comparand]:
        comparand = self.extract_comparand(initial_state)
        zero_comparand = tree_map(jnp.zeros_like, comparand)
        return StochasticState(current_state=initial_state,
                               iterations=0,
                               mean_state=zero_comparand,
                               second_moment_state=zero_comparand)

    def iterate_augmented(self,
                          new_state: State,
                          augmented: StochasticState[State, Comparand]) -> (
                              StochasticState[State, Comparand]):
        def f(value: ComplexNumeric, drift: ComplexNumeric) -> ComplexNumeric:
            return leaky_integrate(value, 1.0, drift, self.convergence_detection_decay,
                                   leaky_average=True)

        mean_state, second_moment_state = self._sufficient_statistics(augmented.current_state)
        new_mean_state = tree_multimap(f, augmented.mean_state, mean_state)
        new_second_moment_state = tree_multimap(f, augmented.second_moment_state,
                                                second_moment_state)
        return StochasticState(current_state=new_state,
                               iterations=augmented.iterations + 1,
                               mean_state=new_mean_state,
                               second_moment_state=new_second_moment_state)

    def converged(self, augmented: StochasticState[State, Comparand]) -> BooleanNumeric:
        data_weight = leaky_data_weight(augmented.iterations, self.convergence_detection_decay)
        mean_squared = tree_map(abs_square, augmented.mean_state)
        return tree_reduce(jnp.logical_and,
                           tree_multimap(partial(jnp.allclose,
                                                 rtol=self.rtol * data_weight,
                                                 atol=self.atol * data_weight),
                                         augmented.second_moment_state,
                                         mean_squared),
                           True)

    def minimum_tolerances(self,
                           augmented: StochasticState[State, Comparand]) -> Tuple[RealNumeric,
                                                                                  RealNumeric]:
        """
        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        data_weight = leaky_data_weight(augmented.iterations, self.convergence_detection_decay)
        mean_squared = tree_map(abs_square, augmented.mean_state)
        variance = tree_multimap(jnp.subtract, augmented.second_moment_state, mean_squared)
        scaled_variance = tree_multimap(divide_nonnegative, variance, mean_squared)

        minimum_atol = divide_nonnegative(tree_reduce(jnp.maximum, tree_map(jnp.amax, variance)),
                                          data_weight)
        minimum_rtol = divide_nonnegative(tree_reduce(jnp.maximum,
                                                      tree_map(jnp.amax, scaled_variance)),
                                          data_weight)
        assert not isinstance(minimum_atol, complex)
        assert not isinstance(minimum_rtol, complex)
        return minimum_atol, minimum_rtol

    # Private methods ------------------------------------------------------------------------------
    def _sufficient_statistics(self, state: State) -> Tuple[Comparand, Comparand]:
        comparand = self.extract_comparand(state)
        squared_comparand = tree_map(abs_square, comparand)
        return comparand, squared_comparand


class StochasticIteratedFunctionWithCombinator(
        IteratedFunctionWithCombinator[Parameters, State, Comparand, Differentiand, Trajectory,
                                       StochasticState[State, Comparand]],
        StochasticIteratedFunction[Parameters, State, Comparand, Trajectory],
        Generic[Parameters, State, Comparand, Differentiand, Trajectory]):
    pass



from numbers import Number
from typing import Any, Optional, Union, overload

import jax.numpy as jnp
import numpy as np

from .annotations import (BooleanNumeric, ComplexArray, ComplexNumeric, IntegralNumeric, RealArray,
                          RealNumeric)

__all__ = ['is_scalar', 'abs_square', 'divide_where', 'divide_nonnegative']


def is_scalar(x: Any) -> bool:
    return isinstance(x, Number) or isinstance(x, (np.ndarray, jnp.ndarray)) and x.shape == ()


@overload
def abs_square(x: ComplexArray) -> RealArray:
    ...


@overload
def abs_square(x: ComplexNumeric) -> RealNumeric:
    ...


def abs_square(x: ComplexNumeric) -> RealNumeric:
    return jnp.square(x.real) + jnp.square(x.imag)


@overload
def divide_where(dividend: RealNumeric,
                 divisor: Union[RealNumeric, IntegralNumeric],
                 *,
                 where: Optional[BooleanNumeric] = None,
                 otherwise: Optional[RealNumeric] = None) -> RealNumeric:
    ...


@overload
def divide_where(dividend: ComplexNumeric,
                 divisor: Union[ComplexNumeric, IntegralNumeric],
                 *,
                 where: Optional[BooleanNumeric] = None,
                 otherwise: Optional[ComplexNumeric] = None) -> ComplexNumeric:
    ...


def divide_where(dividend: ComplexNumeric,
                 divisor: Union[ComplexNumeric, IntegralNumeric],
                 *,
                 where: Optional[BooleanNumeric] = None,
                 otherwise: Optional[ComplexNumeric] = None) -> ComplexNumeric:
    """
    Returns: `jnp.where(where, dividend / divisor, otherwise)`, but without evaluating
        `dividend / divisor` when `where` is false.  This prevents some exceptions.
    """
    if where is None:
        return jnp.true_divide(dividend, divisor)
    dividend = jnp.where(where, dividend, 1.0)
    divisor = jnp.where(where, divisor, 1.0)
    quotient = jnp.true_divide(dividend, divisor)
    return jnp.where(where, quotient, otherwise)


def divide_nonnegative(dividend: RealNumeric, divisor: RealNumeric) -> RealNumeric:
    """
    Returns: The quotient assuming that the dividend and divisor are nonnegative, and infinite
        whenever the divisor equals zero.
    """
    return divide_where(dividend, divisor, where=divisor > 0.0, otherwise=jnp.inf)



from __future__ import annotations

from functools import partial
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar

import jax.numpy as jnp
from jax import jit
from jax.experimental.host_callback import id_tap
from jax.lax import scan, while_loop
from jax.tree_util import tree_multimap

from ..annotations import (BooleanNumeric, IntegralNumeric, PyTree, RealNumeric,
                           TapFunctionTransforms)
from ..dataclasses import dataclass, field
from ..dtypes import default_atol, default_rtol
from .augmented import AugmentedState, State

__all__ = ['IteratedFunction']


Parameters = TypeVar('Parameters', bound=PyTree)
Comparand = TypeVar('Comparand', bound=PyTree)
Trajectory = TypeVar('Trajectory', bound=PyTree)
TheAugmentedState = TypeVar('TheAugmentedState', bound=AugmentedState[Any])
TapFunction = Callable[[None, TapFunctionTransforms], None]


@dataclass
class IteratedFunction(Generic[Parameters, State, Comparand, Trajectory, TheAugmentedState]):
    """
    An IteratedFunction object models an iterated function.

    It is a generic class in terms of five generic types, all of which are pytrees:
        * Parameters, which models the iteration parameters,
        * State, which is the state at each iteration,
        * Comparand, which is calculated from state and compared between iterations to check
          convergence,
        * Trajectory, which is the type produced by sample_trajectory, and
        * TheAugmentedState, which is typically defined by one of the tjax subclasses.

    The two main methods of IteratedFunction are
    * find_fixed_point, which iterates until the fixed point is reached or other stopping conditions
      are met, and
    * sample_trajectory, which iterates for a fixed number of iterations and returns a trajectory.

    Args:
        minimum_iterations: The minimum number of iterations for this fixed point solver.  This must
            be positive.
        maximum_iterations: The maximum number of iterations for this fixed point solver.  This must
            be positive.
        rtol: The relative tolerance for the comparison stopping condition.
        atol: The absolute tolerance for the comparison stopping condition.
    """
    minimum_iterations: IntegralNumeric
    maximum_iterations: IntegralNumeric
    rtol: RealNumeric = field(default_factory=default_rtol)
    atol: RealNumeric = field(default_factory=default_atol)

    # New methods ----------------------------------------------------------------------------------
    def find_fixed_point(self,  # pylint: disable=function-redefined, method-hidden
                         theta: Parameters,
                         initial_state: State) -> TheAugmentedState:
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
        Returns:
            state: The augmented state at the fixed point.
        """
        def f(augmented: TheAugmentedState) -> TheAugmentedState:
            new_state = self.sampled_state(theta, augmented.current_state)
            return self.iterate_augmented(new_state, augmented)

        return while_loop(partial(self.state_needs_iteration, theta),
                          f,
                          self.initial_augmented(initial_state))

    find_fixed_point = jit(find_fixed_point)  # pylint: disable=used-before-assignment

    def sample_trajectory(self,  # pylint: disable=function-redefined, method-hidden
                          theta: Parameters,
                          initial_state: State,
                          maximum_iterations: int,
                          tap_function: Optional[TapFunction]) -> (
                              Tuple[TheAugmentedState, Trajectory]):
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
            maximum_iterations: The number of steps in the trajectory.  Unlike the eponymous member
                variable, this must be static.
        Returns:
            x_star: The augmented state at the fixed point.
            trajectory: A PyTree representing the trajectory of states.
        """
        def f(augmented: TheAugmentedState, x: None) -> Tuple[TheAugmentedState, Trajectory]:
            trajectory: Trajectory
            new_state, trajectory = self.sampled_state_trajectory(theta, augmented)
            new_augmented = self.iterate_augmented(new_state, augmented)
            if tap_function is not None:
                trajectory = id_tap(tap_function, None, result=trajectory)
            return new_augmented, trajectory
        return scan(f, self.initial_augmented(initial_state), None, maximum_iterations)

    # pylint: disable=used-before-assignment
    sample_trajectory = jit(sample_trajectory, static_argnums=(3, 4))

    def debug_fixed_point(self, theta: Parameters, initial_state: State) -> TheAugmentedState:
        """
        This method is identical to find_fixed_point, but avoids using while_loop and its
        concomitant jit.
        """
        augmented = self.initial_augmented(initial_state)
        while self.state_needs_iteration(theta, augmented):
            new_state = self.sampled_state(theta, augmented.current_state)
            augmented = self.iterate_augmented(new_state, augmented)
        return augmented

    def debug_trajectory(self,
                         theta: Parameters,
                         initial_state: State,
                         maximum_iterations: int,
                         tap_function: Optional[TapFunction]) -> (
                             Tuple[TheAugmentedState, Trajectory]):
        """
        This method is identical to sample_trajectory, but avoids using scan and its concomitant
        jit.
        """
        augmented = self.initial_augmented(initial_state)
        for i in range(maximum_iterations):
            trajectory: Trajectory
            concatenated_trajectory: Trajectory
            new_state, trajectory = self.sampled_state_trajectory(theta, augmented)
            augmented = self.iterate_augmented(new_state, augmented)
            concatenated_trajectory = (
                trajectory
                if i == 0
                else tree_multimap(jnp.append, concatenated_trajectory, trajectory))  # noqa: F821
            if tap_function is not None:
                tap_function(None, ())
        return augmented, concatenated_trajectory

    def state_needs_iteration(self, theta: Parameters, augmented: TheAugmentedState) -> bool:
        """
        Args:
            theta: The parameters.
            augmented: The state.
        Returns: True while iteration needs to continue.
        """
        enough_iterations = augmented.iterations >= self.minimum_iterations
        converged = self.converged(augmented)
        not_too_many_iterations = augmented.iterations < self.maximum_iterations
        return jnp.logical_and(not_too_many_iterations,
                               jnp.logical_not(jnp.logical_and(enough_iterations, converged)))

    # Abstract methods -----------------------------------------------------------------------------
    def initial_augmented(self, initial_state: State) -> TheAugmentedState:
        raise NotImplementedError

    def expected_state(self, theta: Parameters, state: State) -> State:
        """
        Returns: The expected value of the next state given the old one.  This is used by the
            combinator.
        """
        raise NotImplementedError

    def sampled_state(self, theta: Parameters, state: State) -> State:
        """
        Returns: A sampled value of the next state in a trajectory.  This is used when finding the
            fixed point.
        """
        raise NotImplementedError

    def sampled_state_trajectory(self,
                                 theta: Parameters,
                                 augmented: TheAugmentedState) -> Tuple[State, Trajectory]:
        """
        Returns:
            sampled_state: A sampled value of the next state in a trajectory.  This is used when
                finding the fixed point.
            trajectory: A value to be concatenated into a trajectory.
        """
        raise NotImplementedError

    def iterate_augmented(self,
                          new_state: State,
                          augmented: TheAugmentedState) -> TheAugmentedState:
        """
        Args:
            new_state: The new state to fold into the augmented state.
            augmented: The last augmented state.
        Returns: The next augmented state.
        """
        raise NotImplementedError

    def converged(self, augmented: TheAugmentedState) -> BooleanNumeric:
        """
        Returns: A Boolean Array of shape () indicating whether the pytrees are close.
        """
        raise NotImplementedError

    def extract_comparand(self, state: State) -> Comparand:
        """
        Returns: The "comparand", which is a subset of the values in the state that are compared by
            subclasses to detect convergence.  In ComparingIteratedFunction, the comparand will be
            compared in successive states to detect convergence.  In StochasticIteratedFunction, the
            mean and variance of the comparand are used to detect convergence.
        """
        raise NotImplementedError

    def minimum_tolerances(self, augmented: TheAugmentedState) -> Tuple[RealNumeric, RealNumeric]:
        """
        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        raise NotImplementedError



from tjax._src.gradient.aliases import adam, adamw
from tjax._src.gradient.chain import ChainedGradientTransformation
from tjax._src.gradient.smd import SMDGradient, SMDState
from tjax._src.gradient.transform import (GradientState, GradientTransformation,
                                          SecondOrderGradientTransformation,
                                          ThirdOrderGradientTransformation)
from tjax._src.gradient.transforms import AdditiveWeightDecay, Scale, ScaleByAdam

__all__ = ['AdditiveWeightDecay', 'ChainedGradientTransformation', 'GradientState',
           'GradientTransformation', 'SMDGradient', 'SMDState', 'Scale', 'ScaleByAdam',
           'SecondOrderGradientTransformation', 'ThirdOrderGradientTransformation', 'adam', 'adamw']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tjax',
 'tjax._src',
 'tjax._src.dataclasses',
 'tjax._src.fixed_point',
 'tjax._src.gradient']

package_data = \
{'': ['*']}

install_requires = \
['colorful>=0.5.4,<0.6.0',
 'jax>=0.2.13,<0.3.0',
 'jaxlib>=0.1',
 'matplotlib>=3.3,<4.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.21',
 'optax>=0,<1',
 'yapf>=0.31']

setup_kwargs = {
    'name': 'tjax',
    'version': '0.12.5',
    'description': 'Tools for JAX.',
    'long_description': "=============\nTools for JAX\n=============\n\n.. role:: bash(code)\n    :language: bash\n\n.. role:: python(code)\n   :language: python\n\nThis repository implements a variety of tools for the differential programming library\n`JAX <https://github.com/google/jax>`_.\n\n----------------\nMajor components\n----------------\n\nTjax's major components are:\n\n- A dataclass decorator :python:`dataclasss` that facilitates defining JAX trees, and has a MyPy plugin.\n  (See `dataclass <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/dataclasses>`_ and `mypy_plugin <https://github.com/NeilGirdhar/tjax/blob/master/tjax/mypy_plugin.py>`_.)\n\n- A fixed point finding library heavily based on `fax <https://github.com/gehring/fax>`_.  Our\n  library (`fixed_point <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/fixed_point>`_):\n\n  - supports stochastic iterated functions,\n  - uses dataclasses instead of closures to avoid leaking JAX tracers, and\n  - supports higher-order differentiation.\n\n----------------\nMinor components\n----------------\n\nTjax also includes:\n\n- An object-oriented wrapper on top of `optax <https://github.com/deepmind/optax>`_.  (See\n  `gradient <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/gradient>`_.)\n\n- A pretty printer :python:`print_generic` for aggregate and vector types, including dataclasses.  (See\n  `display <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/display.py>`_.)\n\n- Versions of :python:`custom_vjp` and :python:`custom_jvp` that support being used on methods.\n  (See `shims <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/shims.py>`_.)\n\n- Tools for working with cotangents.  (See\n  `cotangent_tools <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/cotangent_tools.py>`_.)\n\n- A random number generator class :python:`Generator`.  (See `generator <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/generator.py>`_.)\n\n- JAX tree registration for `NetworkX <https://networkx.github.io/>`_ graph types.  (See\n  `graph <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/graph.py>`_.)\n\n- Leaky integration :python:`leaky_integrate` and Ornstein-Uhlenbeck process iteration\n  :python:`diffused_leaky_integrate`.  (See `leaky_integral <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/leaky_integral.py>`_.)\n\n- An improved version of :python:`jax.tree_util.Partial`.  (See `partial <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/partial.py>`_.)\n\n- A Matplotlib trajectory plotter :python:`PlottableTrajectory`.  (See `plottable_trajectory <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/plottable_trajectory.py>`_.)\n\n- A testing function :python:`assert_jax_allclose` that automatically produces testing code.  And, a related\n  function :python:`jax_allclose`.  (See `testing <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/testing.py>`_.)\n\n- Basic tools like :python:`divide_where`.  (See `tools <https://github.com/NeilGirdhar/tjax/blob/master/tjax/_src/tools.py>`_.)\n\nAlso, see the `documentation <https://neilgirdhar.github.io/tjax/tjax/index.html>`_.\n\n-----------------------\nContribution guidelines\n-----------------------\n\n- Conventions: PEP8.\n\n- How to run tests: :bash:`pytest .`\n\n- How to clean the source:\n\n  - :bash:`isort tjax`\n  - :bash:`pylint tjax`\n  - :bash:`mypy tjax`\n  - :bash:`flake8 tjax`\n",
    'author': 'Neil Girdhar',
    'author_email': 'mistersheik@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NeilGirdhar/tjax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
